from allennlp_models.generation.models.composed_seq2seq import ComposedSeq2Seq
from allennlp_models.generation.models.copynet_seq2seq import CopyNetSeq2Seq
from allennlp_models.generation.models.simple_seq2seq import SimpleSeq2Seq
from allennlp_models.generation.models.bart import Bart
