from allennlp_models.structured_prediction.models.biaffine_dependency_parser import (
    BiaffineDependencyParser,
)
from allennlp_models.structured_prediction.models.constituency_parser import SpanConstituencyParser
from allennlp_models.structured_prediction.models.graph_parser import GraphParser
from allennlp_models.structured_prediction.models.srl import SemanticRoleLabeler
from allennlp_models.structured_prediction.models.srl_bert import SrlBert
